@@ -4,9 +4,9 @@
 use std::collections::{HashMap, HashSet};
 
 use crate::error::{PolarResult, RuntimeError};
-use crate::folder::{fold_term, Folder};
-use crate::formatting::ToPolarString;
+use crate::folder::{fold_list, fold_term, Folder};
 use crate::terms::{has_rest_var, Operation, Operator, Symbol, Term, Value};
+use crate::vm::Goal;
 
 #[derive(Clone, Debug)]
 pub struct Binding(pub Symbol, pub Term);
@@ -40,6 +40,55 @@ pub enum VariableState {
     Partial,
 }
 
+struct Derefer<'a> {
+    binding_manager: &'a BindingManager,
+    seen: HashSet<u64>,
+}
+
+impl<'a> Derefer<'a> {
+    fn new(binding_manager: &'a BindingManager) -> Self {
+        Self {
+            binding_manager,
+            seen: HashSet::new(),
+        }
+    }
+}
+
+impl<'a> Folder for Derefer<'a> {
+    fn fold_list(&mut self, list: Vec<Term>) -> Vec<Term> {
+        let has_rest = has_rest_var(&list);
+        let mut list = fold_list(list, self);
+        if has_rest {
+            let last = list.pop().unwrap();
+            if let Value::List(rest) = last.value() {
+                list.append(&mut rest.clone());
+            } else {
+                list.push(last);
+            }
+        }
+        list
+    }
+
+    fn fold_term(&mut self, t: Term) -> Term {
+        match t.value() {
+            Value::Expression(_) => t,
+            Value::Variable(v) | Value::RestVariable(v) => {
+                let hash = t.hash_value();
+                if self.seen.contains(&hash) {
+                    t
+                } else {
+                    self.seen.insert(hash);
+                    let t = self.binding_manager.lookup(v).unwrap_or(t);
+                    let t = fold_term(t, self);
+                    self.seen.remove(&hash);
+                    t
+                }
+            }
+            _ => fold_term(t, self),
+        }
+    }
+}
+
 /// Represent each binding in a cycle as a unification constraint.
 // TODO(gj): put this in an impl block on VariableState?
 fn cycle_constraints(cycle: Vec<Symbol>) -> Operation {
@@ -104,22 +153,43 @@ impl BindingManager {
         Self::default()
     }
 
+    /// Bind `var` to `val` in the expression `partial`.
+    ///
+    /// If the binding succeeds, the new expression is returned as a goal. Otherwise,
+    /// an error is returned.
+    fn partial_bind(&mut self, partial: Operation, var: &Symbol, val: Term) -> PolarResult<Goal> {
+        match partial.ground(var, val.clone()) {
+            None => Err(RuntimeError::IncompatibleBindings {
+                msg: "Grounding failed".into(),
+            }
+            .into()),
+            Some(grounded) => {
+                self.add_binding(var, val);
+                Ok(Goal::Query {
+                    term: grounded.into_term(),
+                })
+            }
+        }
+    }
+
     // **** State Mutation ***
 
     /// Bind `var` to `val`.
     ///
-    /// If the binding succeeds, Ok is returned. If the binding is *incompatible*
-    /// an error is returned.
+    /// If the binding succeeds, Ok with an optional goal is returned. The goal will be
+    /// present if the binding replaces a partial, which then needs to be reevaluated
+    /// to ensure compatibility.
     ///
-    /// A binding is considered *incompatible* if either:
+    /// If the binding is *incompatible* an error is returned. A binding is considered
+    /// *incompatible* if either:
     ///
     /// 1. `var` is already bound to some value (rebindings are not allowed, even if the
     ///    rebinding is to the same value).
     /// 2. `var` is constrained, and the new binding of `val` is not compatible with those
-    ///    constraints.
+    ///    constraints (as determined by `Operation::ground()`)
     ///
     /// If a binding is compatible, it is recorded. If the binding was to a ground value,
-    /// subsequent calls to `variable_state` or `deref` will return that value.
+    /// subsequent calls to `variable_state` or `deep_deref` will return that value.
     ///
     /// If the binding was between two variables, the two will always have the same value
     /// or constraints going forward. Further, a unification constraint is recorded between
@@ -130,35 +200,36 @@ impl BindingManager {
     ///
     /// If a binding between two variables is made, and one is bound and the other unbound, the
     /// unbound variable will take the value of the bound one.
-    pub fn bind(&mut self, var: &Symbol, val: Term) -> PolarResult<()> {
+    pub fn bind(&mut self, var: &Symbol, val: Term) -> PolarResult<Option<Goal>> {
+        let mut goal = None;
         if let Ok(symbol) = val.value().as_symbol() {
-            self.bind_variables(var, symbol)?;
-        } else if let BindingManagerVariableState::Partial(p) = self._variable_state(var) {
-            if let Some(grounded) = p.ground(var.clone(), val.clone()) {
-                self.add_binding(var, val.clone());
-                self.constrain(&grounded)?;
-            } else {
-                return Err(RuntimeError::IncompatibleBindings {
-                    msg: "Grounding failed".into(),
-                }
-                .into());
-            }
+            goal = self.bind_variables(var, symbol)?;
         } else {
-            if let BindingManagerVariableState::Bound(_) = self._variable_state(var) {
-                return Err(RuntimeError::IncompatibleBindings {
-                    msg: format!("Cannot rebind {:?}", var),
+            match self._variable_state(var) {
+                BindingManagerVariableState::Partial(p) => {
+                    let p = p.clone();
+                    let val = val.clone();
+                    goal = Some(self.partial_bind(p, var, val)?)
                 }
-                .into());
-            }
 
-            self.add_binding(var, val.clone());
+                BindingManagerVariableState::Bound(_) => {
+                    return Err(RuntimeError::IncompatibleBindings {
+                        msg: format!("Cannot rebind {:?}", var),
+                    }
+                    .into())
+                }
+                _ => self.add_binding(var, val.clone()),
+            }
         }
 
         // If the main binding succeeded, the follower binding must succeed.
-        self.do_followers(|_, follower| follower.bind(var, val.clone()))
-            .unwrap();
+        self.do_followers(|_, follower| {
+            follower.bind(var, val.clone())?;
+            Ok(())
+        })
+        .unwrap();
 
-        Ok(())
+        Ok(goal)
     }
 
     /// Rebind `var` to `val`, regardless of compatibility.
@@ -185,37 +256,47 @@ impl BindingManager {
     /// `term` must be an expression`.
     ///
     /// An error is returned if the constraint is incompatible with existing constraints.
-    ///
-    /// (Currently all constraints are considered compatible).
     pub fn add_constraint(&mut self, term: &Term) -> PolarResult<()> {
         self.do_followers(|_, follower| follower.add_constraint(term))?;
 
         assert!(term.value().as_expression().is_ok());
         let mut op = op!(And, term.clone());
+
+        // include all constraints applying to any of its variables.
         for var in op.variables().iter().rev() {
-            match self._variable_state(&var) {
-                BindingManagerVariableState::Unbound => {}
+            match self._variable_state(var) {
                 BindingManagerVariableState::Cycle(c) => {
-                    let mut cycle = cycle_constraints(c);
-                    cycle.merge_constraints(op.clone());
-                    op = cycle;
+                    op = cycle_constraints(c).merge_constraints(op)
                 }
-                BindingManagerVariableState::Partial(e) => {
-                    let mut e = e.clone();
-                    e.merge_constraints(op);
-                    op = e;
-                }
-                BindingManagerVariableState::Bound(v) => {
-                    panic!(
-                        "Unexpected bound variable {var} in constraint. {var} = {val}",
-                        var = var,
-                        val = v
-                    );
+                BindingManagerVariableState::Partial(e) => op = e.clone().merge_constraints(op),
+                _ => {}
+            }
+        }
+
+        let vars = op.variables();
+        let mut varset = vars.iter().collect::<HashSet<_>>();
+
+        // replace any bound variables with their values.
+        for var in vars.iter() {
+            if let BindingManagerVariableState::Bound(val) = self._variable_state(var) {
+                varset.remove(var);
+                match op.ground(var, val) {
+                    Some(o) => op = o,
+                    None => {
+                        return Err(RuntimeError::IncompatibleBindings {
+                            msg: "Grounding failed".into(),
+                        }
+                        .into())
+                    }
                 }
             }
         }
 
-        self.constrain(&op)
+        // apply the new constraint to every remaining variable.
+        for var in varset {
+            self.add_binding(var, op.clone().into_term())
+        }
+        Ok(())
     }
 
     /// Reset the state of `BindingManager` to what it was at `to`.
@@ -234,72 +315,9 @@ impl BindingManager {
     }
 
     // *** Binding Inspection ***
-
-    /// If `term` is a variable, return the value bound to that variable.
-    /// If `term` is a list, dereference all items in the list.
-    /// Otherwise, return `term`.
-    pub fn deref(&self, term: &Term) -> Term {
-        match &term.value() {
-            Value::List(list) => {
-                // Deref all elements.
-                let mut derefed: Vec<Term> =
-                    // TODO(gj): reduce recursion here.
-                    list.iter().map(|t| self.deref(t)).collect();
-
-                // If last element was a rest variable, append the list it derefed to.
-                if has_rest_var(list) {
-                    if let Some(last_term) = derefed.pop() {
-                        if let Value::List(terms) = last_term.value() {
-                            derefed.append(&mut terms.clone());
-                        } else {
-                            derefed.push(last_term);
-                        }
-                    }
-                }
-                term.clone_with_value(Value::List(derefed))
-            }
-            Value::Variable(v) => match self.variable_state(v) {
-                VariableState::Bound(value) => value,
-                _ => term.clone(),
-            },
-            Value::RestVariable(v) => match self.variable_state(v) {
-                VariableState::Bound(value) => match value.value() {
-                    Value::List(l) if has_rest_var(l) => self.deref(&value),
-                    _ => value,
-                },
-                _ => term.clone(),
-            },
-            _ => term.clone(),
-        }
-    }
-
     /// Dereference all variables in term, including within nested structures like
     /// lists and dictionaries.
     pub fn deep_deref(&self, term: &Term) -> Term {
-        pub struct Derefer<'a> {
-            binding_manager: &'a BindingManager,
-        }
-
-        impl<'a> Derefer<'a> {
-            pub fn new(binding_manager: &'a BindingManager) -> Self {
-                Self { binding_manager }
-            }
-        }
-
-        impl<'a> Folder for Derefer<'a> {
-            fn fold_term(&mut self, t: Term) -> Term {
-                match t.value() {
-                    Value::Expression(_) => t,
-                    Value::List(_) => fold_term(self.binding_manager.deref(&t), self),
-                    Value::Variable(_) | Value::RestVariable(_) => {
-                        let derefed = self.binding_manager.deref(&t);
-                        fold_term(derefed, self)
-                    }
-                    _ => fold_term(t, self),
-                }
-            }
-        }
-
         Derefer::new(self).fold_term(term.clone())
     }
 
@@ -411,7 +429,8 @@ impl BindingManager {
 // Private impls.
 impl BindingManager {
     /// Bind two variables together.
-    fn bind_variables(&mut self, left: &Symbol, right: &Symbol) -> PolarResult<()> {
+    fn bind_variables(&mut self, left: &Symbol, right: &Symbol) -> PolarResult<Option<Goal>> {
+        let mut goal = None;
         match (self._variable_state(left), self._variable_state(right)) {
             (
                 BindingManagerVariableState::Bound(left_value),
@@ -491,17 +510,18 @@ impl BindingManager {
             }
             (
                 BindingManagerVariableState::Bound(left_value),
-                BindingManagerVariableState::Partial(_),
+                BindingManagerVariableState::Partial(p),
             ) => {
                 // Left is bound, right has constraints.
-                // TODO (dhatch): No unwrap.
-                self.add_constraint(&op!(Unify, left_value, term!(right.clone())).into_term())?;
+                let p = p.clone();
+                goal = Some(self.partial_bind(p, right, left_value)?);
             }
             (
-                BindingManagerVariableState::Partial(_),
+                BindingManagerVariableState::Partial(p),
                 BindingManagerVariableState::Bound(right_value),
             ) => {
-                self.add_constraint(&op!(Unify, term!(left.clone()), right_value).into_term())?;
+                let p = p.clone();
+                goal = Some(self.partial_bind(p, left, right_value)?);
             }
             (BindingManagerVariableState::Partial(_), _)
             | (_, BindingManagerVariableState::Partial(_)) => {
@@ -511,13 +531,20 @@ impl BindingManager {
             }
         }
 
-        Ok(())
+        Ok(goal)
     }
 
     fn add_binding(&mut self, var: &Symbol, val: Term) {
         self.bindings.push(Binding(var.clone(), val));
     }
 
+    fn lookup(&self, var: &Symbol) -> Option<Term> {
+        match self.variable_state(var) {
+            VariableState::Bound(val) => Some(val),
+            _ => None,
+        }
+    }
+
     /// Look up a variable in the bindings stack and return
     /// a reference to its value if it's bound.
     fn value(&self, variable: &Symbol, bsp: usize) -> Option<&Term> {
@@ -558,22 +585,6 @@ impl BindingManager {
         BindingManagerVariableState::Unbound
     }
 
-    #[allow(clippy::unnecessary_wraps)]
-    fn constrain(&mut self, o: &Operation) -> PolarResult<()> {
-        assert_eq!(o.operator, Operator::And, "bad constraint {}", o.to_polar());
-        for var in o.variables() {
-            match self._variable_state(&var) {
-                // A constraint should not contain a bound variable, it should have been removed in
-                // add_constraint by calling ground.
-                BindingManagerVariableState::Bound(_) => {
-                    panic!("Unexpected bound variable in constraint.")
-                }
-                _ => self.add_binding(&var, o.clone().into_term()),
-            }
-        }
-        Ok(())
-    }
-
     fn do_followers<F>(&mut self, mut func: F) -> PolarResult<()>
     where
         F: FnMut(FollowerId, &mut BindingManager) -> PolarResult<()>,
@@ -589,6 +600,7 @@ impl BindingManager {
 #[cfg(test)]
 mod test {
     use super::*;
+    use crate::formatting::to_polar::ToPolarString;
 
     #[test]
     fn variable_state() {
@@ -772,10 +784,7 @@ mod test {
         let b2 = b1.remove_follower(&b2_id).unwrap();
 
         if let BindingManagerVariableState::Partial(p) = b1._variable_state(&sym!("x")) {
-            assert_eq!(
-                p.to_polar(),
-                "x = y and y = z and y = z and z = x and x > y"
-            );
+            assert_eq!(p.to_polar(), "x = y and y = z and z = x and x > y");
         } else {
             panic!("unexpected");
         }
@@ -788,7 +797,7 @@ mod test {
     }
 
     #[test]
-    fn deref() {
+    fn old_deref() {
         let mut bm = BindingManager::default();
         let value = term!(1);
         let x = sym!("x");
@@ -797,25 +806,25 @@ mod test {
         let term_y = term!(y.clone());
 
         // unbound var
-        assert_eq!(bm.deref(&term_x), term_x);
+        assert_eq!(bm.deep_deref(&term_x), term_x);
 
         // unbound var -> unbound var
         bm.bind(&x, term_y.clone()).unwrap();
-        assert_eq!(bm.deref(&term_x), term_x);
+        assert_eq!(bm.deep_deref(&term_x), term_x);
 
         // value
-        assert_eq!(bm.deref(&value), value.clone());
+        assert_eq!(bm.deep_deref(&value), value.clone());
 
         // unbound var -> value
         let mut bm = BindingManager::default();
         bm.bind(&x, value.clone()).unwrap();
-        assert_eq!(bm.deref(&term_x), value);
+        assert_eq!(bm.deep_deref(&term_x), value);
 
         // unbound var -> unbound var -> value
         let mut bm = BindingManager::default();
         bm.bind(&x, term_y).unwrap();
         bm.bind(&y, value.clone()).unwrap();
-        assert_eq!(bm.deref(&term_x), value);
+        assert_eq!(bm.deep_deref(&term_x), value);
     }
 
     #[test]
