@@ -5,7 +5,7 @@ use std::sync::atomic::{AtomicU64, Ordering};
 
 use crate::bindings::{BindingManager, Bsp, FollowerId, VariableState};
 use crate::counter::Counter;
-use crate::error::PolarResult;
+use crate::error::{PolarError, PolarResult};
 use crate::events::QueryEvent;
 use crate::formatting::ToPolarString;
 use crate::kb::Bindings;
@@ -130,8 +130,7 @@ fn reduce_constraints(bindings: Vec<Bindings>) -> Bindings {
                 .for_each(|(var, value)| match acc.entry(var.clone()) {
                     Entry::Occupied(mut o) => match (o.get().value(), value.value()) {
                         (Value::Expression(x), Value::Expression(y)) => {
-                            let mut x = x.clone();
-                            x.merge_constraints(y.clone());
+                            let x = x.clone().merge_constraints(y.clone());
                             o.insert(value.clone_with_value(value!(x)));
                         }
                         (existing, new) => panic!(
@@ -252,4 +251,8 @@ impl Runnable for Inverter {
     fn clone_runnable(&self) -> Box<dyn Runnable> {
         Box::new(self.clone())
     }
+
+    fn handle_error(&mut self, error: PolarError) -> PolarResult<QueryEvent> {
+        self.vm.handle_error(error)
+    }
 }
