@@ -546,7 +546,7 @@ mod tests {
         let s = r#"
             "this is a \"sub\" string"
         "#;
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         let tok = lexer.next();
         assert!(
             matches!(tok, Some(Ok((_, Token::String(s), _))) if &s == r#"this is a "sub" string"#)
@@ -558,7 +558,7 @@ mod tests {
         let s = r#"
             "💯" 💯
         "#;
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(
             matches!(lexer.next(), Some(Ok((13, Token::String(hunnid), 19))) if hunnid == "💯")
         );
@@ -570,13 +570,13 @@ mod tests {
     #[test]
     fn test_symbol_with_trailing_question_mark() {
         let s = "foo?";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(
             matches!(lexer.next(), Some(Ok((0, Token::Symbol(question), 4))) if question == Symbol::new("foo?"))
         );
 
         let s = "foo??";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         lexer.next();
         assert!(matches!(
             lexer.next(),
@@ -591,7 +591,7 @@ mod tests {
     #[test]
     fn test_symbol_colons() {
         let s = "foo:bar";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(
             matches!(lexer.next(), Some(Ok((0, Token::Symbol(x), 3))) if x == Symbol::new("foo"))
         );
@@ -602,14 +602,14 @@ mod tests {
         assert!(matches!(lexer.next(), None));
 
         let s = "foo::bar";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(
             matches!(lexer.next(), Some(Ok((0, Token::Symbol(x), 8))) if x == Symbol::new("foo::bar"))
         );
         assert!(matches!(lexer.next(), None));
 
         let s = "foo:::bar";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(matches!(
             lexer.next(),
             Some(Err(ParseError::InvalidTokenCharacter {
@@ -623,7 +623,7 @@ mod tests {
     #[test]
     fn test_symbol_question_marks() {
         let s = "foo??";
-        let mut lexer = Lexer::new(&s);
+        let mut lexer = Lexer::new(s);
         assert!(
             matches!(lexer.next(), Some(Ok((0, Token::Symbol(x), 4))) if x == Symbol::new("foo?"))
         );
@@ -642,7 +642,7 @@ mod tests {
     fn test_lexer() {
         let f = r#"hello "world" 12345 < + <= { ] =99 #comment
             more; in; Ruby::Namespace"#;
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(
             matches!(lexer.next(), Some(Ok((0, Token::Symbol(hello), 5))) if hello == Symbol::new("hello"))
         );
@@ -679,39 +679,39 @@ mod tests {
     #[allow(clippy::float_cmp)]
     fn test_numbers() {
         let f = "1+2";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(lexer.next(), Some(Ok((0, Token::Integer(1), 1)))));
         assert!(matches!(lexer.next(), Some(Ok((1, Token::Add, 2)))));
         assert!(matches!(lexer.next(), Some(Ok((2, Token::Integer(2), 3)))));
 
         let f = "0123";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(
             lexer.next(),
             Some(Ok((0, Token::Integer(123), 4)))
         ));
 
         let f = "1.ee1";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(
             lexer.next(),
             Some(Err(ParseError::InvalidFloat { .. }))
         ));
 
         let f = "1.1";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(lexer.next(), Some(Ok((_, Token::Float(f), _))) if f == 1.1));
 
         let f = "1e1";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(lexer.next(), Some(Ok((_, Token::Float(f), _))) if f == 1e1));
 
         let f = "1e-1";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(lexer.next(), Some(Ok((_, Token::Float(f), _))) if f == 1e-1));
 
         let f = "1.1e-1";
-        let mut lexer = Lexer::new(&f);
+        let mut lexer = Lexer::new(f);
         assert!(matches!(lexer.next(), Some(Ok((_, Token::Float(f), _))) if f == 1.1e-1));
     }
 }
